/*******************************************************************************
 * Copyright (c) 2010, 2012 Tasktop Technologies
 * Copyright (c) 2010, 2011 SpringSource, a division of VMware
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     Tasktop Technologies - initial API and implementation
 ******************************************************************************/
package com.tasktop.c2c.server.scm.service;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.tasktop.c2c.server.common.service.EntityNotFoundException;
import com.tasktop.c2c.server.common.service.ValidationException;
import com.tasktop.c2c.server.common.service.WrappedCheckedException;
import com.tasktop.c2c.server.common.service.domain.Region;
import com.tasktop.c2c.server.common.service.web.AbstractVersionedRestServiceClient;
import com.tasktop.c2c.server.scm.domain.Blame;
import com.tasktop.c2c.server.scm.domain.Blob;
import com.tasktop.c2c.server.scm.domain.BranchAccessConfig;
import com.tasktop.c2c.server.scm.domain.Commit;
import com.tasktop.c2c.server.scm.domain.CommitReference;
import com.tasktop.c2c.server.scm.domain.Diff;
import com.tasktop.c2c.server.scm.domain.Import;
import com.tasktop.c2c.server.scm.domain.Merge;
import com.tasktop.c2c.server.scm.domain.Profile;
import com.tasktop.c2c.server.scm.domain.RawBlob;
import com.tasktop.c2c.server.scm.domain.RevPathItem;
import com.tasktop.c2c.server.scm.domain.ScmLastVisited;
import com.tasktop.c2c.server.scm.domain.ScmRepository;
import com.tasktop.c2c.server.scm.domain.ScmRepositoryACL;
import com.tasktop.c2c.server.scm.domain.ScmRepositoryEvent;
import com.tasktop.c2c.server.scm.domain.ScmSummary;
import com.tasktop.c2c.server.scm.domain.Search;
import com.tasktop.c2c.server.scm.domain.Trees;
import com.tasktop.c2c.server.scm.domain.WikiBlob;
import com.tasktop.c2c.server.scm.service.requests.InitRepositoryRequest;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;

/**
 * A client to communicate to the SCM REST service.
 * <p/>
 * This client object should be configured with an RestClientDelegate, e.g. ApacheHttpRestClientDelegate, which is
 * provided in SDK, to handle HTTP REST communication.
 * <p/>
 * Usage:
 * 
 * <pre>
 * ScmServiceClient client = new ScmServiceClient();
 * ApacheHttpRestClientDelegate delegate = new ApacheHttpRestClientDelegate(user, passwd);
 * client.setRestClientDelegate(delegate);
 * client.setBaseUrl(REST_API_URL);
 * </pre>
 */
public class ScmServiceClient extends AbstractVersionedRestServiceClient implements ScmService {

	/**
	 * Essentially a map entry, but spring does not like map return results that are meant to be response data.
	 * 
	 */
	public static final class CommitsForAuthor {
		private Profile author;
		private Integer count;

		public void setAuthor(Profile author) {
			this.author = author;
		}

		public Profile getAuthor() {
			return author;
		}

		public void setCount(Integer count) {
			this.count = count;
		}

		public Integer getCount() {
			return count;
		}
	}

	@SuppressWarnings("unused")
	// All of the setters in this method are used programmatically by the JSON serializer.
	private static class ServiceCallResult {
		private List<Commit> commitList;
		private List<CommitsForAuthor> commitsForAuthorList;
		private ScmRepository scmRepository;
		private List<ScmRepository> scmRepositoryList;
		private List<ScmSummary> scmSummaryList;
		private Commit commit;
		private String string;
		private Trees trees;
		private Blob blob;
		private WikiBlob wikiBlob;
		private Blame blame;
		private RevPathItem item;
		// private List<DiffEntry> diffEntries;
		private Diff diff;
                private Merge.Result result;
                private Search.Result searchResult;
                private List<ScmRepositoryEvent> scmRepositoryEventList;
                private BranchAccessConfig branchACL;

		public List<Commit> getCommitList() {
			return commitList;
		}

		public void setCommitList(List<Commit> commitList) {
			this.commitList = commitList;
		}

		public List<CommitsForAuthor> getCommitsForAuthorList() {
			return commitsForAuthorList;
		}

		public void setCommitsForAuthorList(List<CommitsForAuthor> commitsForAuthorList) {
			this.commitsForAuthorList = commitsForAuthorList;
		}

		public ScmRepository getScmRepository() {
			return scmRepository;
		}

		public void setScmRepository(ScmRepository scmRepo) {
			this.scmRepository = scmRepo;
		}

		public List<ScmRepository> getScmRepositoryList() {
			return scmRepositoryList;
		}

		public void setScmRepositoryList(List<ScmRepository> scmRepoSet) {
			this.scmRepositoryList = scmRepoSet;
		}

		public List<ScmSummary> getScmSummaryList() {
			return scmSummaryList;
		}

		public void setScmSummaryList(List<ScmSummary> scmSummaryList) {
			this.scmSummaryList = scmSummaryList;
		}

		public Commit getCommit() {
			return commit;
		}

		public void setCommit(Commit commit) {
			this.commit = commit;
		}

		public String getString() {
			return string;
		}

		public void setString(String string) {
			this.string = string;
		}

		public Trees getTrees() {
			return trees;
		}

		public void setTrees(Trees trees) {
			this.trees = trees;
		}

		public Blob getBlob() {
			return blob;
		}

                public WikiBlob getWikiBlob() {
                        return wikiBlob;
                }


		public void setBlob(Blob blob) {
			this.blob = blob;
		}

		public Blame getBlame() {
			return blame;
		}

		public void setBlame(Blame blame) {
			this.blame = blame;
		}

                public BranchAccessConfig getBranchACL() {
                    return branchACL;
                }

                public void setBranchACL(BranchAccessConfig branchACL) {
                    this.branchACL = branchACL;
                }

		public RevPathItem getRevPathItem() {
			return item;
		}

		public void setRevPathItem(RevPathItem item) {
			this.item = item;
		}

//		private List<DiffEntry> getDiffEntryList() {
//			return diffEntries;
//		}
//
//		public void setDiffEntryList(List<DiffEntry> diffEntries) {
//			this.diffEntries = diffEntries;
//		}

		private Diff getDiff() {
			return diff;
		}

		public void setDiff(Diff diff) {
			this.diff = diff;
		}

                public Merge.Result getResult() {
                    return result;
                }

                public void setResult(Merge.Result mergeResult) {
                    this.result = mergeResult;
                }

                public Search.Result getSearchResult() {
                    return searchResult;
                }

                public void setSearchResult(Search.Result searchResult) {
                    this.searchResult = searchResult;
                }
                
                public List<ScmRepositoryEvent> getScmRepositoryEventList() {
                    return scmRepositoryEventList;
                }

                public void setScmRepositoryEventList(List<ScmRepositoryEvent> scmRepositoryEventList) {
                    this.scmRepositoryEventList = scmRepositoryEventList;
                }


	}

	private abstract class GetCall<T> {

		public abstract T getValue(ServiceCallResult result);

		public T doCall(String urlStub, Object... variables) {
			ServiceCallResult callResult = getServiceCallResult(urlStub, variables);

			T retVal = getValue(callResult);

			if (retVal == null) {
				throw new IllegalStateException("Illegal result from call to scmService");
			}

			return retVal;
		}

                protected ServiceCallResult getServiceCallResult(String urlStub, Object[] variables) {
                    return delegate.getForObject(computeUrl(urlStub), ServiceCallResult.class,
                            variables);
                }
	}

	private abstract class PostCall<T> {

		public abstract T getValue(ServiceCallResult result);

		public T doCall(String urlStub, Object objToPost, Object... variables) {
			ServiceCallResult callResult = getServiceCallResult(urlStub,objToPost, variables);

			T retVal = getValue(callResult);

			if (retVal == null) {
				throw new IllegalStateException("Illegal result from call to scmService");
			}

			return retVal;
		}
                
                protected ServiceCallResult getServiceCallResult(String urlStub, Object objToPost, Object[] variables) {
                    return delegate.postForObject(computeUrl(urlStub), objToPost,
                            ServiceCallResult.class, variables);
                }
	}
        
        private abstract class DeleteCall {
		
		public void doCall(String urlStub, Object objToPost, Object... variables) {
			delegate.deleteForObject(computeUrl(urlStub), objToPost, ServiceCallResult.class, variables);
		}
                
	}

	public static final String GET_SCM_REPOSITORIES_URL = "repository";

	public List<ScmRepository> getScmRepositories() throws EntityNotFoundException {

		try {
			return new GetCall<List<ScmRepository>>() {
				public List<ScmRepository> getValue(ServiceCallResult result) {
					return result.getScmRepositoryList();
				}
			}.doCall(GET_SCM_REPOSITORIES_URL);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
			throw e;
		}
	}

	public static final String CREATE_SCM_REPOSITORY_URL = "repository";

	public ScmRepository createScmRepository(ScmRepository scmRepository) throws EntityNotFoundException,
			ValidationException {
		try {
			return new PostCall<ScmRepository>() {
				public ScmRepository getValue(ServiceCallResult result) {
					return result.getScmRepository();
				}
			}.doCall(CREATE_SCM_REPOSITORY_URL, scmRepository);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
			convertValidationException(e);
			throw e;
		}
	}

	public static final String REPOSITORY_ID_URLPARAM = "repositoryId";

	public static final String DELETE_SCM_REPOSITORY_URL = "repository/delete";

	public void deleteScmRepository(ScmRepository repo) throws EntityNotFoundException {
		try {
			new PostCall<Object>() {

				@Override
				public Object getValue(ServiceCallResult result) {
					return "empty";
				}

			}.doCall(DELETE_SCM_REPOSITORY_URL, repo);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
			throw e;
		}
	}

	public static final String OFFSET_URL_PARAM = "offset";
	public static final String PAGESIZE_URL_PARAM = "pageSize";
	public static final String NUMDAYS_URL_PARAM = "numDays";

	public static final String GET_LOG_URL = "log";

	public List<Commit> getLog(Region region) {

		// Calculate the correct URL now.
		String url = GET_LOG_URL;
		if (region != null) {
			url = String.format("%s?%s=%s&%s=%s", url, OFFSET_URL_PARAM, region.getOffset(), PAGESIZE_URL_PARAM,
					region.getSize());
		}

		return new GetCall<List<Commit>>() {
			public List<Commit> getValue(ServiceCallResult result) {
				return result.getCommitList();
			}
		}.doCall(url);
	}

	public static final String GET_NUM_COMMITS_BY_AUTHOR_URL = "commitsByAuthor";

	public Map<Profile, Integer> getNumCommitsByAuthor(int numDays) {
		String url = String.format("%s?%s=%s", GET_NUM_COMMITS_BY_AUTHOR_URL, NUMDAYS_URL_PARAM, numDays);

		return toMap(new GetCall<List<CommitsForAuthor>>() {
			public List<CommitsForAuthor> getValue(ServiceCallResult result) {
				return result.getCommitsForAuthorList();
			}
		}.doCall(url));
	}

	private Map<Profile, Integer> toMap(List<CommitsForAuthor> commitsForAuthor) {
		Map<Profile, Integer> result = new HashMap<Profile, Integer>();
		for (CommitsForAuthor c : commitsForAuthor) {
			result.put(c.getAuthor(), c.getCount());
		}
		return result;
	}

	public static final String GET_SCM_SUMMARY_URL = "summary";

	public List<ScmSummary> getScmSummary(int numDays) {
		String url = String.format("%s?%s=%s", GET_SCM_SUMMARY_URL, NUMDAYS_URL_PARAM, numDays);

		return new GetCall<List<ScmSummary>>() {
			@Override
			public List<ScmSummary> getValue(ServiceCallResult result) {
				return result.getScmSummaryList();
			}
		}.doCall(url);
	}

	public static final String GET_COMMIT_URL = "commit/{repoName}/{commitId}";

	public Commit getCommit(String repoName, String commitId, Integer context) throws EntityNotFoundException {

		try {
			String url = GET_COMMIT_URL;
			if (context != null) {
				url = String.format("%s?%s=%s", url, "lctx", context);
			}

			return new GetCall<Commit>() {
				@Override
				public Commit getValue(ServiceCallResult result) {
					return result.getCommit();
				}
			}.doCall(url, repoName, commitId);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
		}
		throw new IllegalStateException();
	}

        public static final String DO_COMMIT_URL = "commit/{repoName:.*}";

        public Commit doCommit(String repoName, Merge.Request commitRequest)
                throws EntityNotFoundException {

            String url = GET_BLOB_URL;

            return new PostCall<Commit>() {
                public Commit getValue(ServiceCallResult result) {
                        return result.getCommit();
                }
            }.doCall(url, commitRequest);
        }


	public static final String GET_DIFF_ENTRIES_URL = "diffentries/{repoName}/{baseCommitId}/{commitId}";

	public Diff getDiffEntries(String repoName, String baseCommitId, String commitId, Integer numContextLines)
			throws EntityNotFoundException {
		try {
			return new GetCall<Diff>() {
				@Override
				public Diff getValue(ServiceCallResult result) {
					return result.getDiff();
				}
			}.doCall(GET_DIFF_ENTRIES_URL, repoName, baseCommitId, commitId);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
		}
		throw new IllegalStateException();
	}


        public static final String GET_DIFF_URL = "diff/{repoName}";

        public Diff getDiff(String repoName, String revSpec, Integer numContextLines, String file)
			throws EntityNotFoundException {
		try {
			return new GetCall<Diff>() {
				@Override
				public Diff getValue(ServiceCallResult result) {
					return result.getDiff();
				}
			}.doCall(GET_DIFF_URL, repoName, revSpec);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
		}
		throw new IllegalStateException();
	}

	public static final String GET_LOG_FOR_REPO_URL = "repository/{repo}/repolog";

	public List<Commit> getLog(String repoName, Region region) {
		// Calculate the correct URL now.
		String url = GET_LOG_FOR_REPO_URL;
		if (region != null) {
			url = String.format("%s?%s=%s&%s=%s", url, OFFSET_URL_PARAM, region.getOffset(), PAGESIZE_URL_PARAM,
					region.getSize());
		}

		return new GetCall<List<Commit>>() {
			public List<Commit> getValue(ServiceCallResult result) {
				return result.getCommitList();
			}
		}.doCall(url, repoName);
	}

	public static final String GET_LOG_FOR_PATH_URL = "repository/{repo}/commits/{revision}";

	public List<Commit> getLog(String repoName, String revision, String path, Region region)
			throws EntityNotFoundException {
            // Calculate the correct URL now.
            String url = GET_LOG_FOR_PATH_URL;
            if (region != null) {
                url = String.format("%s/%s?%s=%s&%s=%s", url, path, OFFSET_URL_PARAM, region.getOffset(),
                        PAGESIZE_URL_PARAM, region.getSize());
            }

            try {
                return new GetCall<List<Commit>>() {
                    public List<Commit> getValue(ServiceCallResult result) {
                        return result.getCommitList();
                    }
                }.doCall(url, repoName, revision);
            } catch (WrappedCheckedException e) {
                convertEntityNotFoundException(e);
            }
            throw new IllegalStateException();
	}

	public static final String START_REVISION = "startRevision";
	public static final String END_REVISION = "endRevision";
	public static final String DOT3 = "dot3";
        public static final String GET_LOG_FOR_RANGE_URL = "repository/{repo}/rangelog";
        public static final String GET_LOG_FOR_RANGE_URL_WITH_PATH = String.format("%s/{path}", GET_LOG_FOR_RANGE_URL);
        public List<Commit> getLogForRange(final String repoName, final String startRevision,
            final String endRevision, final String path, final boolean dot3, final Region region) throws EntityNotFoundException {
            if (repoName == null) {
                throw new IllegalArgumentException("No repository name.");
            } else if (startRevision == null) {
                throw new IllegalArgumentException("No startRevision.");
            }

            //prepare request parameters
            final Map<String, Object> requestParams = new LinkedHashMap<String, Object>();
            requestParams.put(START_REVISION, startRevision);//mandatory
            requestParams.put(END_REVISION, endRevision);
            requestParams.put(DOT3, dot3);
            requestParams.put(OFFSET_URL_PARAM, region != null ? region.getOffset() : null);
            requestParams.put(PAGESIZE_URL_PARAM, region != null ? region.getSize() : null);

            //prepare URL variables
            final Collection<Object> variables = new ArrayList<Object>();
            variables.add(repoName);
            if (path != null) {
                variables.add(path);
            }
            variables.addAll(requestParams.values());
            variables.removeAll(Collections.singleton(null));

            //prepare url template
            final StringBuilder urlStub = new StringBuilder();
            for (final Map.Entry<String, Object> paramEntry : requestParams.entrySet()) {
                if (urlStub.length() == 0) {
                    urlStub.append((path != null) ? GET_LOG_FOR_RANGE_URL_WITH_PATH : GET_LOG_FOR_RANGE_URL);
                    appendFirstRequestParameter(urlStub, paramEntry.getKey());
                } else if (paramEntry.getValue() != null) {
                    appendNextRequestParameter(urlStub, paramEntry.getKey());
                }
            }

            try {
                return new GetCall<List<Commit>>() {
                    @Override
                    public List<Commit> getValue(ServiceCallResult result) {
                        return result.getCommitList();
                    }
                }.doCall(urlStub.toString(), variables.toArray(new Object[variables.size()]));
            } catch (WrappedCheckedException e) {
                convertEntityNotFoundException(e);
            }
            throw new IllegalStateException();
    }

    private void appendNextRequestParameter(final StringBuilder urlStub, final String parameterName) {
        appendRequestParameter(urlStub, parameterName, "&");
    }

    private void appendFirstRequestParameter(final StringBuilder urlStub, final String parameterName) {
        appendRequestParameter(urlStub, parameterName, "?");
    }

    private void appendRequestParameter(final StringBuilder urlStub, String parameterName, String separator) {
        urlStub.append(separator).append(parameterName).append("={").append(parameterName).append("}");
    }

	public static final String BRANCH_PARAM = "branch";
	public static final String GET_LOG_FOR_BRANCH_URL = "repository/{repo}/branchlog";

	public List<Commit> getLogForBranch(String repoName, String branchName, Region region) {
		// Calculate the correct URL now.
		String url = GET_LOG_FOR_BRANCH_URL;
		if (region != null) {
			url = String.format("%s?%s=%s&%s=%s&%s=%s", url, BRANCH_PARAM, branchName, OFFSET_URL_PARAM,
					region.getOffset(), PAGESIZE_URL_PARAM, region.getSize());
		}

		return new GetCall<List<Commit>>() {
			public List<Commit> getValue(ServiceCallResult result) {
				return result.getCommitList();
			}
		}.doCall(url, repoName);
	}

	public static final String BRANCH_URL = "branch/{repoName}/{branchName}";

	/** @Override in fact it does override, but ... */
	public String createBranch(String repoName, String base, String branchName) throws EntityNotFoundException {
		try {
                        Merge.Request mr = new Merge.Request();
                        mr.setSrc(base);


			return new PostCall<String>() {
				public String getValue(ServiceCallResult result) {
					return result.getString();
				}
			}.doCall(BRANCH_URL, mr, repoName, branchName);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
			// convertValidationException(e);
			throw e;
		}
	}

        /** @Override in fact it does override, but ... */
	public String renameBranch(String repoName, String branchName, String newName ) throws EntityNotFoundException {
		try {
                        Merge.Request mr = new Merge.Request();
                        mr.setTarget(newName);

			return new PostCall<String>() {
				public String getValue(ServiceCallResult result) {
					return result.getString();
				}
			}.doCall(BRANCH_URL, mr, repoName, branchName);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
			// convertValidationException(e);
			throw e;
		}
	}

	/** @Override in fact it does override, but ... */
	public void deleteBranch(String repoName, String branchName) throws EntityNotFoundException {
		try {
			new DeleteCall() {}.doCall(BRANCH_URL, "", repoName, branchName);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
			// convertValidationException(e);
			throw e;
		}
	}

        /** @Override in fact it does override, but ... */

        public static final String TAG_URL = "tag/{repoName}/{tagName}";

	public String createTag(String repoName, String base, String branchName) throws EntityNotFoundException {
		try {
                        Merge.Request mr = new Merge.Request();
                        mr.setSrc(base);

			return new PostCall<String>() {
				public String getValue(ServiceCallResult result) {
					return result.getString();
				}
			}.doCall(TAG_URL, mr, repoName, branchName);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
			// convertValidationException(e);
			throw e;
		}
	}

        public String renameTag(String repoName, String tagName, String newName) throws EntityNotFoundException {
		try {
                        Merge.Request mr = new Merge.Request();
                        mr.setTarget(newName);

			return new PostCall<String>() {
				public String getValue(ServiceCallResult result) {
					return result.getString();
				}
			}.doCall(TAG_URL, mr, repoName, tagName);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
			// convertValidationException(e);
			throw e;
		}
	}

	/** @Override in fact it does override, but ... */
	public void deleteTag(String repoName, String branchName) throws EntityNotFoundException {
		try {
			new DeleteCall() {}.doCall(TAG_URL, "", repoName, branchName);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
			// convertValidationException(e);
			throw e;
		}
	}

	public static final String PUBILC_SSH_KEY_URL = "sshkey";

	public String getPublicSshKey() {
		return new GetCall<String>() {
			public String getValue(ServiceCallResult result) {
				return result.getString();
			}
		}.doCall(PUBILC_SSH_KEY_URL);
	}

	public String getClientVersion() {
		return ScmService.VERSION;
	}

	public static final String REVISION_PARAM = "revision";

	/*
	 * This does not work. sadly public static final String GET_TREES_URL =
	 * "repository/{repo}/trees/{revision}/{filePath:[.*]/*}";
	 */
	public static final String GET_TREES_URL = "repository/{repo}/trees/{revision}";

	/** @Override in fact it does override, but ... */
	public Trees getTrees(String repo, String revision, String path, boolean history, int recursion)
			throws EntityNotFoundException {
		// Calculate the correct URL now.
		String url = GET_TREES_URL;

		url = String.format("%s/%s?%s=%b&%s=%d", url, path, "history", history, "recursion", recursion);

		try {
			return new GetCall<Trees>() {
				public Trees getValue(ServiceCallResult result) {
					return result.getTrees();
				}
			}.doCall(url, repo, revision);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
		}
		throw new IllegalStateException();

	}

	public static final String GET_BLOB_URL = "repository/{repo}/blob/{revision}";

	/** @Override in fact it does override, but ... */
	public Blob getBlob(String repository, String revision, String path) throws EntityNotFoundException {

		String url = GET_BLOB_URL;

		url = String.format("%s/%s", url, path);

		try {
			return new GetCall<Blob>() {
				public Blob getValue(ServiceCallResult result) {
					return result.getBlob();
				}
			}.doCall(url, repository, revision);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
		}
		throw new IllegalStateException();

	}

        public static final String GET_RAWBLOB_URL = "repository/{repo}/raw/{revision}";
        /** @Override in fact it does override, but ... */
	public RawBlob getRawBlob(String repository, String revision, String path) throws EntityNotFoundException {
            throw new UnsupportedOperationException( "Raw blobs are for internal use oly");
        }


        public static final String GET_WIKIBLOB_URL = "repository/{repo}/wiki/{revision}";

        /** @Override in fact it does override, but ... */
	public WikiBlob getWikiBlob(String repository, String revision, String path, boolean source) throws EntityNotFoundException {

		String url = GET_BLOB_URL;

		url = String.format("%s/%s?source=%s", url, path, source);

		try {
			return new GetCall<WikiBlob>() {
				public WikiBlob getValue(ServiceCallResult result) {
					return result.getWikiBlob();
				}
			}.doCall(url, repository, revision);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
		}
		throw new IllegalStateException();

	}

        public String previewWikiBlob(String repoName, String revision, String path, Blob blob)
                throws EntityNotFoundException {

            String url = GET_BLOB_URL;

            url = String.format("%s/%s", url, path);

            return new PostCall<String>() {
                public String getValue(ServiceCallResult result) {
                        return result.getString();
                }
            }.doCall(url, blob);
        }


	public static final String GET_BLAME_URL = "repository/{repo}/blame/{revision}";

	/** @Override in fact it does override, but ... */
	public Blame getBlame(String repository, String revision, String path) throws EntityNotFoundException {

		String url = GET_BLAME_URL;

		url = String.format("%s/%s", url, path);

		try {
			return new GetCall<Blame>() {
				public Blame getValue(ServiceCallResult result) {
					return result.getBlame();
				}
			}.doCall(url, repository, revision);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
		}
		throw new IllegalStateException();

	}

	public static final String GET_ITEM_URL = "repository/{repo}/item/{revision}";

	/** @Override in fact it does override, but ... */
	public RevPathItem getItem(String repository, String revision, String path) throws EntityNotFoundException {

		String url = GET_ITEM_URL;

		url = String.format("%s/%s", url, path);

		try {
			return new GetCall<RevPathItem>() {
				public RevPathItem getValue(ServiceCallResult result) {
					return result.getRevPathItem();
				}
			}.doCall(url, repository, revision);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
		}
		throw new IllegalStateException();

	}

	public static final String REV_A_PARAM = "revA";
	public static final String REV_B_PARAM = "revB";
	public static final String GET_MERGE_BASE_URL = "repository/{repo}/merge-base";

	/** @Override in fact it does override, but ... */
	public Commit getMergeBase(String repository, String revA, String revB) throws EntityNotFoundException {

		String url = String.format("%s?%s={revA}&%s={revB}", GET_MERGE_BASE_URL, REV_A_PARAM, REV_B_PARAM);

		try {
			return new GetCall<Commit>() {
				public Commit getValue(ServiceCallResult result) {
					return result.getCommit();
				}
			}.doCall(url, repository, revA, revB);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
		}
		throw new IllegalStateException();

	}

	public static final String TARGET_BRANCH = "target";
	public static final String SRC_BRANCH = "src";
        public static final String MERGE_URL = "repository/{repo}/merge";
        public Merge.Result merge(String repository, String target, String src) throws EntityNotFoundException {
                String url = String.format("%s?%s={target}&%s={src}", MERGE_URL, TARGET_BRANCH, SRC_BRANCH);

		try {
			return new GetCall<Merge.Result>() {
				public Merge.Result getValue(ServiceCallResult result) {
					return result.getResult();
				}
			}.doCall(url, repository, target, src);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
		}
		throw new IllegalStateException();

        }

        public Merge.Result merge(Merge.Request request) throws EntityNotFoundException, ValidationException {
            try {
			return new PostCall<Merge.Result>() {
				public Merge.Result getValue(ServiceCallResult result) {
					return result.getResult();
				}
			}.doCall(MERGE_URL, request, request.getRepository());
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);
			convertValidationException(e);
			throw e;
		}
        }

        public static final String INDEX_URL = "repository/{repo}/index";

        public void indexScmRepository(String repo) throws EntityNotFoundException {
            try {
                    new PostCall<Object>() {
                            public Object getValue(ServiceCallResult result) {
                                    return "indexing";
                            }
                    }.doCall(INDEX_URL, repo);
		} catch (WrappedCheckedException e) {
			convertEntityNotFoundException(e);			
			throw e;
		}
        }

        public static final String SEARCH_URL = "repository/{repo}/search";
        public static final String SEARCH_REVISION_URL_PARAM = "revision";
        public static final String SEARCH_PHRASE_URL_PARAM = "q";
        public static final String SEARCH_TYPES_URL_PARAM = "t";
        public static final String SEARCH_PATH_URL_PARAM = "p";

        public Search.Result search( String repository, String revision, String phrase, String types, String path, Region region ) throws EntityNotFoundException {
            try {
                    String url = SEARCH_URL;
                    if (region != null) {
                            // XXX Pther params missing
                            url = String.format("%s?%s=%s&%s=%s", url, OFFSET_URL_PARAM, region.getOffset(), PAGESIZE_URL_PARAM,
                                            region.getSize());
                    }
                    return new GetCall<Search.Result>() {
                            public Search.Result getValue(ServiceCallResult result) {
                                    return result.getSearchResult();
                            }
                    }.doCall(url, repository, revision, phrase, types);
            } catch (WrappedCheckedException e) {
                    convertEntityNotFoundException(e);
            }
            throw new IllegalStateException();
        }

        public static final String GET_SCM_REPOSITORY_URL = "repository/{repo}/retrieve";

        public ScmRepository getRepositoryByName(String repoName) throws EntityNotFoundException {
            try {
                return new GetCall<ScmRepository>() {
                    @Override
                    public ScmRepository getValue(ServiceCallResult result) {
                        return result.getScmRepository();
                    }
                }.doCall(GET_SCM_REPOSITORY_URL, repoName);
            } catch (WrappedCheckedException e) {
                convertEntityNotFoundException(e);
                throw e;
            }
        }


        public static final String LAST_REPO_URL = "lrc";

        public ScmLastVisited getLastVisited() {
            // ToDo
            throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
        }

        public void setLastVisited(ScmLastVisited lv) {
            // ToDo
            throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
        }
        
        
        public static final String DESCRIPTION_REPO_URL = "repository/description";
        
        public void setRepositoryDescription(ScmRepository scmRepository) throws EntityNotFoundException, ValidationException {
            throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
        }
        
        public static final String INIT_REPO_URL = "repository/init";
        
        /**
         * Initializes empty repository by creating README.md file, ONLY hosted repositories.
         * @param request
         * @throws EntityNotFoundException 
         */
        public void initRepository(InitRepositoryRequest request) {
            //TODO
            throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
        }

        public static final String IMPORT_REPOSITORY_URL = "import/{repo:.*}";

        public Import.Status importRepository( String repoName, Import.Request request ) {
            // TODO
            throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
        }

        public Import.Status importRepositoryStatus( String repoName ) {
            // TODO
            throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
        }

        public static final String GET_REPOSITORY_EVENTS_URL = "repository/events";

	public List<ScmRepositoryEvent> getRepositoryEvents(Region region) {

		// Calculate the correct URL now.
		String url = GET_REPOSITORY_EVENTS_URL;
		if (region != null) {
			url = String.format("%s?%s=%s&%s=%s", url, OFFSET_URL_PARAM, region.getOffset(), PAGESIZE_URL_PARAM,
					region.getSize());
		}

		return new GetCall<List<ScmRepositoryEvent>>() {
			public List<ScmRepositoryEvent> getValue(ServiceCallResult result) {
				return result.getScmRepositoryEventList();
			}
		}.doCall(url);
	}
        
    public static final String REPOSITORY_PROTECTED_BRANCHES = "repository/acl";
    
    public List<ScmRepositoryACL> getScmRepositoriesACL() {
        throw new UnsupportedOperationException("Not supported yet.");
    }
    
    public static final String ACL_FOR_BRANCH_URL = "repository/{repo}/acl/{branch:.*}";
        
    public BranchAccessConfig getBranchAccessConfig(String repo, String branch) {
        try {
            return new GetCall<BranchAccessConfig>() {
                @Override
                public BranchAccessConfig getValue(ServiceCallResult result) {
                    return result.getBranchACL();
                }

                @Override
                protected ServiceCallResult getServiceCallResult(String urlStub, Object[] variables) {
                    final BranchAccessConfig acl = delegate.getForObject(computeUrl(urlStub), BranchAccessConfig.class,
                            variables);
                    final ServiceCallResult result = new ServiceCallResult();
                    result.setBranchACL(acl);
                    return result;
                }
                
            }.doCall(ACL_FOR_BRANCH_URL, repo, branch);
        } catch (WrappedCheckedException e) {
        }
        throw new IllegalStateException();
    }

    public BranchAccessConfig setBranchAccessConfig(String repo, String branch, final BranchAccessConfig acl) {
        try {
            return new PostCall<BranchAccessConfig>() {
                @Override
                public BranchAccessConfig getValue(ServiceCallResult result) {
                    return acl;
                }

                @Override
                protected ServiceCallResult getServiceCallResult(String urlStub, Object objToPost, Object[] variables) {
                    final BranchAccessConfig acl = delegate.postForObject(computeUrl(urlStub), objToPost,
                            BranchAccessConfig.class, variables);
                    final ServiceCallResult result = new ServiceCallResult();
                    result.setBranchACL(acl);
                    return result;
                }

            }.doCall(ACL_FOR_BRANCH_URL, acl, repo, branch);
        } catch (WrappedCheckedException e) {
        }
        throw new IllegalStateException();
    }

    public static final String CREATE_COMMIT_REFERENCE_URL = "reference/create";
    public void updateCommitReference(final CommitReference commitReference) {
        try {
            new PostCall<Object>() {
                public Object getValue(final ServiceCallResult result) {
                    return "";
                }
            }.doCall(CREATE_COMMIT_REFERENCE_URL, commitReference);
        } catch (WrappedCheckedException e) {
            throw e;
        }
    }

    public static final String CREATE_COMMIT_REFERENCES_URL = "references/create";
    public void updateCommitReferences(List<CommitReference> commitReferences) {
        try {
            new PostCall<Object>() {
                public Object getValue(final ServiceCallResult result) {
                    return "";
                }
            }.doCall(CREATE_COMMIT_REFERENCES_URL, commitReferences);
        } catch (WrappedCheckedException e) {
            throw e;
        }
    }

    
    public List<CommitReference> getCommitReferences(final String repository, String pathPrefix) {
        throw new UnsupportedOperationException("Not supported yet."); 
    }

}
